C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfgkvalk (crec, kval, idbg, ierr)

C @(#)Subroutine: rjfgkvalk          Author: R.A. Jansen -- Oct  3, 2009
C @(#)
C @(#)  Purpose: Read an integer-valued keyword value 'kval' from a FITS
C @(#)    header record 'crec', already known to contain an integer-val-
C @(#)    ued keyword. 'kval' is declared as a variable of the INTEGER*8
C @(#)    (64-bit integer) data type.   If the keyword value was written
C @(#)    in free format (allowed in v3.0 of the FITS standard),  'crec'
C @(#)    will be modified  and returned in fixed format  upon exit.  If
C @(#)    verbosity level 'idbg'>0,  additional messages may be printed.
C @(#)    Should any error be detected, error status 'ierr' will be non-
C @(#)    zero upon exit.  Its value  (if >2000)  uniquely identifies  a
C @(#)    known error condition.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL rjfgkvalk (crec, kval, idbg, ierr),   where  'crec'  is a
C @(#)    char string with a declared length of 80 characters and 'kval'
C @(#)    is an integer of the INTEGER*8 data type.  'idbg'  and  'ierr'
C @(#)    are both integers of the default INTEGER data type.
C @(#)
C @(#)  Note: this task is meant to be run after 'rjfvkval' (when 'ityp'
C @(#)    has the return value of 5) and 'rjfgkwrd'.
C @(#)
C @(#)  Dependencies:	clen, cinit, cidx, lenc
C @(#)
C @(#)  Revision history:
C @(#)    Oct  3, 2009 [RAJ] -- Initial version of this routine.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80
      INTEGER*8 kval
      INTEGER	idbg, ierr

C --- Declaration of local variables...
      CHARACTER ctmp*80, cval*80, dbid*19, cc
      INTEGER	clen, lenc, cidx, ic, i, j, k, m
      INTEGER*8 ktmp

C     'ib','ie','i0','i9','ip','im' and 'is' are shorthand integer equi-
C     valents in the ASCII collating sequence of space, equals sign,
C     digits 0 and 9, plus and minus, and slash, respectively...
      INTEGER	ib, ie, i0, i9, ip, im, is

C     Flag 'lsgn' is raised if a + or - sign is found...
      LOGICAL	lsgn

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfgkvalk): '

C --- Initialize the output variable...
      kval  = 0

C --- Initialize local variables (space, =, 0, 9, +, -, / )...
      CALL cinit(ctmp)
      CALL cinit(cval)
      ib = ICHAR(' ')
      ie = ICHAR('=')
      i0 = ICHAR('0')
      i9 = ICHAR('9')
      ip = ICHAR('+')
      im = ICHAR('-')
      is = ICHAR('/')
      lsgn = .FALSE.

C --- Test that byte 9 in 'crec' is indeed an equals sign (captures user
C     error of calling 'rjfgkvalk' before 'rjfvkval')...
      cc = crec(9:9)
      ic = ICHAR(cc)
      IF (ic.NE.ie) THEN
C        Not an equals-sign, therefore no keyword value in this record!
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'header record does not contain a keyword value'
         ierr = 2661
         GOTO 9
      ENDIF

C --- Assign the nominal value portion of the header record to 'ctmp'.
C     In v3.0 of the FITS standard, free-format values are allowed to
C     start anywhere after character 10 ...
      ctmp(1:71) = crec(10:80)


C     The following counters store: j = actual length (position of last
C     non-blank character) of 'ctmp', i = position of first non-blank
C     character in 'ctmp'...
      j = clen(ctmp)
      i = lenc(ctmp)


C --- Test that the first non-blank character in the value portion of
C     the record is either a sign or a digit (captures the user error of
C     calling 'rjfgkvalk' while the return value of 'rjfvkval' != 5) ...
      cc = ctmp(i:i)
      ic = ICHAR(cc)
      IF (ic.NE.ip .AND. ic.NE.im .AND. (ic.LT.i0.OR.ic.GT.i9) ) THEN
C        Unexpected first non-blank char; keyword not integer-valued!
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value is not an '//
     +      'integer (does not start with a sign or digit)'
         ierr = 2662
         GOTO 9
      ELSE IF (ic.EQ.ip .OR. ic.EQ.im) THEN
C        First non-blank character is a sign character...
         IF (idbg.GT.2) WRITE(*,*) dbid//'  keyword value starts '//
     +      'with a sign'
         lsgn = .TRUE.
      ENDIF


C --- Copy the non-blank characters in 'ctmp', one character at a time,
C     to temporary value string 'cval', while testing on-the-fly for any
C     illegal character that is inconsistent with an integer value.  The
C     first char in 'cval' is blank to aid in the list-directed read ...
      m = 2
      cval(m:m) = cc

C  >> Begin-of-loop over the contiguous characters in 'ctmp', starting
C     with the 2nd non-blank character, until we find a blank or slash..
      k = i
1     k = k + 1

      cc = ctmp(k:k)
      ic = ICHAR(cc)

C     Test whether this is a valid character...
      IF ((ic.LT.i0 .OR. ic.GT.i9) .AND. ic.NE.ip .AND. ic.NE.im
     +    .AND. ic.NE.ib .AND. ic.NE.is                          ) THEN
C        Illegal character: neither a sign nor a digit...
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value contains '//
     +      'character other than sign or digits'
         ierr = 2663
         GOTO 9
      ELSE IF (lsgn .AND. (ic.EQ.ip .OR. ic.EQ.im) ) THEN
C        More than one sign detected: value is not an integer...
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value contains '//
     +      'more than one sign: not an integer'
         ierr = 2664
         GOTO 9
      ELSE IF (ic.EQ.ip .OR. ic.EQ.im) THEN
C        Embedded sign detected: value is not an integer...
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value contains '//
     +      'spurious embedded sign: not an integer'
         ierr = 2665
         GOTO 9
      ELSE IF (ic.EQ.ib .OR. ic.EQ.is) THEN
C        We have read all characters representing the integer value...
         cc = ' '
      ENDIF

C     Character is indeed valid, so copy it to 'cval'...
      m = m + 1
      cval(m:m) = cc

C     Decide whether to continue testing the next character...
      IF (ic.NE.ib.AND.ic.NE.is.AND.k.LT.j) GOTO 1

C  << End-of-loop.  Counter i may be re-used.


C --- Test whether we exited the loop because of an embedded space...
      IF (ic.EQ.ib .AND. k.LT.j) THEN
C        Indeed, we exited due to a space; next non-blank character, if
C        any, must be a / ...
2        k = k + 1
         cc = ctmp(k:k)
         ic = ICHAR(cc)
         IF (ic.EQ.ib .AND. k.LT.j) GOTO 2

         IF (k.LT.j .AND. ic.NE.ib .AND. ic.NE.is) THEN
C           Detected a non-blank character other than a / ...
            IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted keyword '//
     +         'value with embedded space character(s)'
            ierr = 2666
            GOTO 9
         ENDIF
      ENDIF


C --- Try to read the integer keyword value from string 'cval'...
      READ(cval,*,IOSTAT=ierr) ktmp
      IF (ierr.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'failed to read INTEGER*8 keyword value from record'
         ierr = 2667
         GOTO 9
      ENDIF
      IF (ktmp.LT.-9.2233720368547E18
     +    .OR. ktmp.GT.9.2233720368547E18) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//' keyword value read from '//
     +      'record is out of range for INTEGER*8'
         ierr = 2668
         GOTO 9
      ENDIF
C     Return the integer value (INTEGER*8 data type)...
      kval = ktmp


C --- Return a possibly modified version of input header record 'crec',
C     with the keyword value (and comment) in fixed format...

C     First test whether the start position of any keyword comment needs
C     to be adjusted...
      k  = cidx(crec((m+10):80),'/')
      k  = k + (m+9)
      IF (k.GT.32) THEN
C        Shift keyword comment to the left...
         crec(32:(32+(80-k))) = crec(k:80)
         DO i = (32+(80-k)+1),80,1
            crec(i:i) = ' '
         ENDDO
      ELSE IF (k.LT.32) THEN
C        Shift keyword comment to the right...
         crec(32:80) = crec(k:(80-(32-k)))
         DO i = k,31,1
            crec(i:i) = ' '
         ENDDO
      ENDIF
C     Then right-justify the keyword value itself on byte 30 ...
      crec(11:31) = '                     '
      crec((33-m):30) = cval(2:(m-1))


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, kval=',kval,
     +   ', ierr=',ierr

      RETURN
      END


