C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfgkvals (crec, sval, ilen, idbg, ierr)

C @(#)Subroutine: rjfgkvals          Author: R.A. Jansen -- Sep 24, 2009
C @(#)
C @(#)  Purpose:  Read  a character string-valued  keyword value  'sval'
C @(#)    from a FITS header record 'crec',  already known  to contain a
C @(#)    string-valued keyword.  The opening and closing quotes will be
C @(#)    stripped and any embedded pair of single quotes is replaced by
C @(#)    a single quote.  The actual  (rather than declared)  length of
C @(#)    the keyword value is returned through  'ilen',  where trailing
C @(#)    spaces will be ignored but leading ones deemed significant. In
C @(#)    the special case of a null-string (''), the returned length is
C @(#)    -1  to distinguish it from a blank string (' ' or '        '),
C @(#)    which has a length of 0.   If the keyword value was written in
C @(#)    free-format (allowed in v3.0 of the FITS standard) 'crec' will
C @(#)    be modified and returned in fixed format on exit. If verbosity
C @(#)    level 'idbg'>0, additional messages may be printed. Should any
C @(#)    error be detected, error status  'ierr'  will be non-zero upon
C @(#)    exit.  Its value (if >2000) uniquely identifies  a known error
C @(#)    condition.
C @(#)
C @(#)  Calling sequence: CALL rjfgkvals (crec, sval, ilen, idbg, ierr),
C @(#)    where 'crec' and 'sval' are char strings with declared lengths
C @(#)    of 80-characters, and 'ilen', 'idbg' and 'ierr' are integers.
C @(#)
C @(#)  Note: this task is meant to be run after 'rjfvkval' (when 'ityp'
C @(#)    has the return value of 1) and 'rjfgkwrd'.
C @(#)
C @(#)  Dependencies:	cinit, clen, cidx, lenc, ctrim
C @(#)
C @(#)  Revision history:
C @(#)    Jun 23, 2009 [RAJ] -- Initial draft version of this routine.
C @(#)    Jun 28, 2009 [RAJ] -- Changed calling sequence  to return also
C @(#)                          the length of the "value" itself.
C @(#)    Sep 23, 2009 [RAJ] -- Major redink;  changed calling sequence;
C @(#)    --Sep 24, 2009        added tests; return input header record
C @(#)                          in fixed-format.
C @(#)    Oct  2, 2009 [RAJ] -- Fixed bug in adjusting a keyword comment
C @(#)                          when returning 'crec' in fixed format.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER*80 crec, sval
      INTEGER	ilen, idbg, ierr

C --- Declaration of local variables...
      CHARACTER	ctrim*80, ctmp*80, cval*80, dbid*19, cc
      INTEGER	clen, lenc, cidx, ic, i, j, k, m, nq, ll

C     'ib','ie','iq','is' and 'it' are shorthand integer equivalents in
C     the ASCII collating sequence of the space, equals sign, quote,
C     slash and tilde, respectively...
      INTEGER	ib, ie, iq, is, it
      CHARACTER	cq

C     Flag 'lclos' is raised when the closing quote is found...
      LOGICAL	lclos

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid = 'DEBUG(rjfgkvals): '

C --- Initialize the output variables...
      CALL cinit(sval)
      ilen = -1

C --- Initialize the local variables (space, slash, tilde, quote)...
      CALL cinit(ctmp)
      CALL cinit(cval)
      ib = ICHAR(' ')
      ie = ICHAR('=')
      iq = 39
      is = ICHAR('/')
      it = ICHAR('~')
      cq = CHAR(iq)
      lclos = .FALSE.

C --- Test that byte 9 in 'crec' is indeed an equals sign (captures user
C     error of calling 'rjfgkvals' before 'rjfvkval')...
      cc = crec(9:9)
      ic = ICHAR(cc)
      IF (ic.NE.ie) THEN
C        Not an equals-sign, therefore no keyword value in this record!
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'header record does not contain a keyword value'
         ierr = 2631
         GOTO 9
      ENDIF

C --- Assign the nominal value portion of the header record to 'ctmp'.
C     In v3.0 of the FITS standard, free-format values are allowed to
C     start anywhere after character 10 ...
      ctmp(1:71) = crec(10:80)

C     The following counters store: i = actual length (position of last
C     non-blank character) of 'ctmp', j = position of first non-blank
C     character in 'ctmp', k = position of the first quote in 'ctmp'...
      i = clen(ctmp)
      j = lenc(ctmp)
      k = cidx(ctmp,cq)

C --- Test that the first non-blank character in the value portion of
C     the record was indeed a single quote (captures the user error of
C     calling 'rjfgkvals' while the return value of 'rjfvkval' !=1) ...
      IF (j.NE.k) THEN
C        First non-blank character is not a quote, not string-valued!
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value is not a '//
     +      'character string (does not start with a single quote)'
         ierr = 2632
         GOTO 9
      ENDIF


C --- Copy the character string value to 'cval' and 'sval', one char at
C     a time, while searching for the closing quote.  Replace embedded
C     pairs of quotes by a single quote in 'sval', and test on-the-fly
C     for any illegal characters outside the restricted ASCII set...

      cval(1:1) = cq
      nq = 0
      m  = j

C  >> Loop over the characters in 'ctmp', starting with the character
C     that follows the opening quote...
1     m = m + 1
      cc = ctmp(m:m)
      ic = ICHAR(cc)

C     Test whether this characters is within the restricted ASCII set...
      IF (ic.LT.ib.OR.ic.GT.it) THEN
C        Illegal character outside the restricted ASCII set!
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'keyword value contains non-printable character'
         CALL cinit(sval)
         ierr = 2633
         GOTO 9
      ENDIF

C     Character is indeed valid, so copy it to 'cval'...
      cval((m-j+1):(m-j+1)) = cc

C     Test whether the character is a quote...
      IF (ic.EQ.iq) THEN
C        We have found a quote...
         IF (m.EQ.i) THEN
C           The closing quote is the final character in the record...
            IF (idbg.GT.1) WRITE(*,*) dbid//' detected closing quote'
            lclos = .TRUE.
         ELSE
C           Test whether it is followed by a second quote...
            cc = ctmp((m+1):(m+1))
            ic = ICHAR(cc)
            IF (cc.EQ.cq) THEN
C              We found an embedded pair of quotes; skip to next char...
               IF (idbg.GT.1) WRITE(*,*) dbid//
     +            ' detected embedded pair of quotes'
               nq = nq + 1
               cval((m-j+2):(m-j+2)) = cc
               sval((m-j):(m-j)) = cc
               m = m + 1
            ELSE
C              We appear to have found the closing quote; test the next
C              character (should only be blank or possibly a slash)...
               IF (ic.EQ.ib.OR.ic.EQ.is) THEN
C                 We have indeed found the closing quote...
                  IF (idbg.GT.1) WRITE(*,*) dbid//
     +               ' detected closing quote'
                  lclos = .TRUE.
               ELSE
C                 Invalid character following apparent closing quote...
                  IF (idbg.GT.0) WRITE(*,*) dbid//'malformatted stri'//
     +               'ng value: invalid character following apparent '//
     +               'closing quote'
                  CALL cinit(sval)
                  ierr = 2634
                  GOTO 9
               ENDIF
            ENDIF
         ENDIF
      ELSE
C        Not a quote; copy the character to 'sval'...
         sval((m-j-nq):(m-j-nq)) = cc
      ENDIF

C     Decide whether to continue testing the next character...
      IF ((.NOT.lclos).AND.m.LT.71) GOTO 1

C     Test whether we exited the loop due to an end-of-string condition
C     (signaling a missing closing quote)...
      IF (.NOT.lclos) THEN
         IF (idbg.GT.0) WRITE(*,*)
     +      dbid//'missing closing quote in character string value'
         ierr = 2635
         GOTO 9
      ENDIF

C  << End-of-loop.  

C     Preserve values of m and j; counters i and k may be reused...


C --- Strip any trailing spaces in excess of the minimum string-length
C     of 8 characters from 'cval' (but capture the special cases of the
C     null and empty strings). Return a possibly modified version of the
C     input header record 'crec' in fixed format...
      ctmp = ctrim(cval,cq,0)
      i = clen(ctmp)

C     Padd short strings with spaces to a minimum of 8 characters, but
C     leave a null string ((m-j)=1) unchanged...
      IF (i.LT.8.AND.(i.GT.0.OR.(m-j).GT.1)) i = 8
      CALL cinit(cval)
      cval = cq//ctmp(1:i)//cq
      ll = clen(cval)
      crec(11:(m+9)) = cval(1:(m-1))


C --- See if the starting position of any keyword comment needs to be
C     adjusted to the right or left...
      i = MAX(32,(10+ll+2))
      k = cidx(crec((m+10):80),'/')
      k = k + (m+9)
      IF (k.GT.i) THEN
         crec(i:(i+(80-k))) = crec(k:80)
         DO m = (i+(80-k)+1),80,1
            crec(m:m) = ' '
         ENDDO
      ELSE IF (k.LT.i) THEN
         crec(i:80) = crec(k:(80-(i-k)))
         DO m = k,(i-1),1
            crec(m:m) = ' '
         ENDDO
      ENDIF


C --- Return the actual length 'ilen' of the keyword value 'sval'...
      IF ((m-j).GT.1) ilen = clen(sval)


9     IF (idbg.GT.2) THEN
         WRITE(*,*) dbid//crec(1:clen(crec))
         WRITE(*,*) dbid//'  cval="'//cval(1:ll)//'", ll=',ll
         WRITE(*,*) dbid//'  sval="'//sval(1:ilen)//'", ilen=',ilen
      ENDIF
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


