C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfpcomm (crec, ckey, ccom, idbg, ierr)

C @(#)Subroutine: rjfpcomm           Author: R.A. Jansen -- Jul 10, 2009
C @(#)
C @(#)  Purpose:  Write a comment record to a FITS header record  'crec'
C @(#)    consisting of an optional keyword  'ckey' (which may be blank,
C @(#)    or one of the  reserved keywords  "COMMENT" or "HISTORY")  and
C @(#)    the comment text  'ccom'.  The comment text will be written to
C @(#)    the header record without quotes  (unless these are explicitly
C @(#)    present in 'ccom' on input, and with leading blanks considered
C @(#)    to be significant.  Character 9 of 'crec' will be blank, while
C @(#)    the comment text will be written starting from char 10. If the
C @(#)    comment text exceeds 71 characters, it will be truncated.  The
C @(#)    comment should only contain characters in the restricted ASCII
C @(#)    set.  If verbosity level 'idbg'>0,  additional messages may be
C @(#)    printed.  Error status 'ierr' will be non-zero on exit, should
C @(#)    any error condition be detected. Its value (if >2000) uniquely
C @(#)    identifies a known error condition.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL rjfpcomm(crec, ckey, ccom, idbg, ierr),  where 'crec' and
C @(#)    'ccom' are strings with a declared lengths of 80 chars, 'ckey'
C @(#)    is a string with a declared length of 8 characters, and 'idbg'
C @(#)    and 'ierr' are both integers.
C @(#)
C @(#)  Note: It is not an error to invoke this routine with both 'ckey'
C @(#)    and 'ccom' a blank string,  although it is far more  efficient
C @(#)    to just run 'cinit' in that case.
C @(#)
C @(#)  Dependencies:	clen, cinit, ucase
C @(#)
C @(#)  Revision history:
C @(#)    Jul 10, 2009 [RAJ] -- Initial version of this routine.
C @(#)    Sep 21, 2009 [RAJ] -- Modified check for restricted ASCII set.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80, ckey*8, ccom*80
      INTEGER	idbg, ierr

C --- Declaration of local variables...
      CHARACTER dbid*19, cc
      INTEGER	clen, klen, ic, j

C     'iA','iZ','i0','i9','ih','iu','ib' and 'it' are shorthand integer
C     equivalents in the ASCII collating sequence of uppercase letters A
C     and Z, numbers 0 and 9, hyphen, underscore, space, and tilde ...
      INTEGER	iA, iZ, i0, i9, ih, iu, ib, it

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfpcomm): '

C --- Initialize the output variable...
      CALL cinit(crec)
 
C --- Initialize the local variables...
      iA = ICHAR('A')
      iZ = ICHAR('Z')
      i0 = ICHAR('0')
      i9 = ICHAR('9')
      ih = ICHAR('-')
      iu = ICHAR('_')
      ib = ICHAR(' ')
      it = ICHAR('~')

C --- If non-blank, initialize the output FITS record with an uppercase
C     version of the keyword...
      klen = clen(ckey)
      IF (klen.NE.0) THEN

         CALL ucase(ckey,crec)
         ckey(1:8) = crec(1:8)
      
C        Test whether the 1st char in the keyword starts with an upper
C        case letter...
         cc = ckey(1:1)
         ic = ICHAR(cc)
         IF (ic.LT.iA.OR.ic.GT.iZ) THEN
C           First character of keyword is not a letter of the alphabet..
            IF (idbg.GT.0) WRITE(*,*)
     +         dbid//'comment keyword does not start with a letter'
            ierr = 2331
            GOTO 9
         ENDIF

C        Test that characters 2--8 comply with the standard...
         DO j=2,klen,1
            cc = ckey(j:j)
            ic = ICHAR(cc)
            IF ( (ic.GE.iA.AND.ic.LE.iZ) .OR. (ic.GE.i0.AND.ic.LE.i9)
     +          .OR. ic.EQ.ih .OR. ic.EQ.iu ) THEN
C              character is compliant with the FITS standard...
               CONTINUE
            ELSE IF (ic.EQ.ib) THEN
C              keyword has invalid embedded blank character(s)...
               IF (idbg.GT.0) WRITE(*,*) dbid//'invalid keyword with'//
     +            ' embedded blank(s)'
               ierr = 2332
               GOTO 9
            ELSE
C              keyword contains invalid non-printable character(s)...
               IF (idbg.GT.0) WRITE(*,*) dbid//'invalid keyword with'//
     +            ' embedded non-printable character(s)'
               ierr = 2333
               GOTO 9
            ENDIF
         ENDDO
      ENDIF


C --- Test on-the-fly for any illegal characters outside the restricted
C     ASCII set while copying verbatim the (possibly truncated) comment
C     text to 'crec'...
      klen = MIN(clen(ccom),71)

C     The following loop must not be executed unless klen>0 ...
      DO j=1,klen
         cc = ccom(j:j)
         ic = ICHAR(cc)
         IF (ic.LT.ib.OR.ic.GT.it) THEN
C           Illegal character outside the restricted ASCII set!
            IF (idbg.GT.0) WRITE(*,*) dbid//'comment text contains '//
     +         'embedded non-printable character(s)'
            ierr = 2334
            GOTO 9
         ENDIF
         crec((9+j):(9+j)) = cc
      ENDDO


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


