C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfpkvale (crec, rval, cfmt, idbg, ierr)

C @(#)Subroutine: rjfpkvale          Author: R.A. Jansen -- Jul 13, 2009
C @(#)
C @(#)  Purpose:  Write a real-valued keyword value to a FITS header re-
C @(#)    cord 'crec',  for a keyword value 'rval' that is declared as a
C @(#)    variable of the default REAL data type.   For most host OS and
C @(#)    compiler implementations this corresponds to a 32-bit (REAL*4)
C @(#)    floating point real.  The keyword value will be written right-
C @(#)    justified  on byte 30 in the header record  using the  Fortran
C @(#)    format specification 'cfmt'.  If blank on input,  default for-
C @(#)    matting will be used.   Only  "Fw.d" and "Ew.dEe"  formats are
C @(#)    allowed ("D" or "G" are not), "w" must not exceed 20, opening/
C @(#)    closing parentheses and/or quotes must not be included in this
C @(#)    specification,  nor should any Fortran scaling offset  ("kP").
C @(#)    Valid format specifications are, eg: 'F6.4', 'F20.1', 'F20.7',
C @(#)    'E10.4','E20.1','E20.7','E20.7e2','E20.7e3',' '. A prefix "SP"
C @(#)    or "+" to the format string will force printing "+" as well as
C @(#)    "-" signs.  The entire format specifier must not exceed the 8-
C @(#)    character limit of 'cfmt', however.  If verbosity level 'idbg'
C @(#)    >0,  additional messages may be printed.   Error status 'ierr'
C @(#)    will be non-zero upon exit,  should any error condition be de-
C @(#)    tected. Its value (if >2000) uniquely identifies a known error
C @(#)    condition.
C @(#)
C @(#)  Calling sequence:  CALL rjfpkvale(crec, rval, cfmt, idbg, ierr),
C @(#)    where  'crec'  is a character string with a declared length of
C @(#)    80 characters,  'rval'  is a variable of the default REAL data
C @(#)    type,  and 'cfmt' a character string with a declared length of
C @(#)    8 characters.  'idbg' and 'ierr' are both integers.
C @(#)
C @(#)  Note1: 32-bit floating point REALs generally represent at most 6
C @(#)    or 7 significant digits faithfully;  this should be taken into
C @(#)    account when choosing the format specification 'cfmt'.
C @(#)  Note2: Internally, scaling offset 1P is enforced for exponential
C @(#)    notation, such that a value of pi is written as  3.1415927E+00
C @(#)    instead of the Fortran default 0.31415927E+01. This 1P (or any
C @(#)    other P modifier) must *not* be part of 'cfmt' on invokation.
C @(#)  Note3: this task is meant to write a real keyword value in fixed
C @(#)    or exponential format to a header record that has been initia-
C @(#)    lized with the FITS keyword;  use one of 'rjfmkval?' to modify
C @(#)    an existing keyword value.
C @(#)
C @(#)  Dependencies:	ucase, cidx, cldx, clen, cinit, lenc
C @(#)
C @(#)  Revision history:
C @(#)    Jul 13, 2009 [RAJ] -- Initial version of this routine.
C @(#)    Oct  6, 2009 [RAJ] -- Modified code to strip all but a single 
C @(#)                          trailing 0 if 'cfmt' was blank on input.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80, cfmt*8
      REAL	rval
      INTEGER	idbg, ierr

C --- Declaration of local variables...
      CHARACTER dbid*19, cval*80, sfmt*16, cc
      REAL	rtmp, rdev
      INTEGER	clen, cidx, cldx, lenc, ic, i, j, k
      LOGICAL	lexp

C     The following variables are shorthand integer equivalents in the
C     ASCII collating sequence of the digits 0 and 9, and the space, E,
C     F, S, P, +, -, and . , respectively...
      INTEGER	i0, i9, ib, iEE, iFF, iPP, iSS, ip, im, id

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfpkvale): '

C --- Initialize local variables...
      CALL cinit(cval)
      i0  = ICHAR('0')
      i9  = ICHAR('9')
      ib  = ICHAR(' ')
      iEE = ICHAR('E')
      iFF = ICHAR('F')
      iPP = ICHAR('P')
      iSS = ICHAR('S')
      ip  = ICHAR('+')
      im  = ICHAR('-')
      id  = ICHAR('.')
      rtmp= ABS(rval)
      lexp= .FALSE.

C --- Make sure the format specification uses only uppercase letters...
      CALL ucase(cfmt,sfmt)
      cfmt = sfmt(1:8)
      CALL cinit(sfmt)


C --- Test whether the format specification was left blank...
      k = clen(cfmt)
      IF (k.EQ.0) THEN
C        Blank format specification: use default formatting...
         IF (idbg.GT.1) WRITE(*,*) dbid//' using default formatting'
         IF (rtmp.GT.1.0D-3.AND.rtmp.LT.1.0D3) THEN
            IF (rtmp.LT.1.0D-2) THEN
               cfmt = 'F20.9'
            ELSE IF (rtmp.LT.1.0D-1) THEN
               cfmt = 'F20.8'
            ELSE IF (rtmp.LT.1.0D1) THEN
               cfmt = 'F20.7'
            ELSE IF (rtmp.LT.1.0D2) THEN
               cfmt = 'F20.6'
            ELSE
               cfmt = 'F20.5'
            ENDIF
         ELSE IF (rtmp.LT.1.0D100) THEN
            cfmt = 'E20.7E2'
         ELSE
            cfmt = 'E20.7E3'
         ENDIF
      ENDIF


C --- Test the apparent validity of the format specification:
C     There must not appear any parentheses in the format specifier...
      i = cidx(cfmt,'(')
      IF (i.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not contain opening parenthesis'
         ierr = 2381
      ENDIF
      i = cldx(cfmt,')')
      IF (i.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not contain closing parenthesis'
         ierr = 2382
      ENDIF
C     There must not appear any scaling offset in the format specifier,
C     but there may be an "SP" sign-modifier...
      i = cidx(cfmt,'P')
      j = cldx(cfmt,'S')
      IF (i.NE.0.AND.j.EQ.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not have scaling offset (P)'
         ierr = 2383
      ENDIF
      IF (i.NE.0.AND.j.NE.0) THEN
         IF (j.NE.(i-1)) THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier contains "S" and "P" but not as is "SP"'
            ierr = 2384
         ENDIF
      ENDIF
      j = cldx(cfmt,'P')
      IF (i.NE.j) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not have scaling offset (P)'
         ierr = 2383
      ENDIF
C     Double precision values are not allowed in the present routine...
      i = cidx(cfmt,'D')
      IF (i.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not be double precision (D)'
         ierr = 2385
      ENDIF
C     Generalized format specifications are not supported at present...
      i = cidx(cfmt,'G')
      IF (i.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not be generalized (G)'
         ierr = 2386
      ENDIF
C     There must be one and only one decimal point in the format...
      i = cidx(cfmt,'.')
      IF (i.EQ.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must contain a decimal point (.)'
         ierr = 2387
      ENDIF
      j = cldx(cfmt,'.')
      IF (i.NE.j) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must have only one decimal point (.)'
         ierr = 2388
      ENDIF
C     The two characters bracketting the decimal point must be digits...
      cc = cfmt((i-1):(i-1))
      ic = ICHAR(cc)
      IF (ic.LT.i0.OR.ic.GT.i9) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier has incorrect syntax'
         ierr = 2389
      ENDIF
      cc = cfmt((i+1):(i+1))
      ic = ICHAR(cc)
      IF (ic.LT.i0.OR.ic.GT.i9) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier has incorrect syntax'
         ierr = 2390
      ENDIF
C     Format specification must be "F"ixed or "E"xponential notation...
      i = cidx(cfmt,'E')
      j = cldx(cfmt,'F')
      IF (i.EQ.0.AND.j.EQ.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must contain either F or E format'
         ierr = 2391
      ELSE IF (i.NE.0.AND.j.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not contain both F and E'
         ierr = 2392
      ELSE IF (i.NE.0.AND.j.EQ.0) THEN
         lexp = .TRUE.
      ENDIF
      i = cidx(cfmt,'F')
      IF (i.NE.j) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier must not contain more than one F'
         ierr = 2393
      ENDIF
C     The minimum length of any valid format specification is 4...
      i = clen(cfmt)
      IF (i.LT.4) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'format specifier too short to contain a valid format'
         ierr = 2394
      ENDIF
C     Check for other illegal characters within the format specifier...
      DO i=1,8
         cc = cfmt(i:i)
         ic = ICHAR(cc)
         IF (     ic.NE.ib.AND.ic.NE.id
     +       .AND.ic.NE.iEE.AND.ic.NE.iFF
     +       .AND.ic.NE.iSS.AND.ic.NE.iPP
     +       .AND.ic.NE.ip.AND.(ic.LT.i0.OR.ic.GT.i9) ) THEN
            IF (idbg.GT.0) WRITE(*,*) dbid//
     +         'detected illegal characters within format specifier'
            ierr = 2395
         ENDIF
      ENDDO
C     Return if any of the above tests failed...
      IF (ierr.NE.0) GOTO 9


C --- Build the internal format string from the requested and verified
C     format specifier, then try to write the real-valued keyword value
C     to character string 'cval'...
      j = clen(cfmt)
      cc = cfmt(1:1)
      ic = ICHAR(cc)
      IF (lexp) THEN
         IF (ic.EQ.ip) THEN
            sfmt = '(1PSP'//cfmt(1:j)//')'
         ELSE
            sfmt = '(1P'//cfmt(1:j)//')'
         ENDIF
      ELSE
         IF (ic.EQ.ip) THEN
            sfmt = '(SP'//cfmt(1:j)//')'
         ELSE
            sfmt = '('//cfmt(1:j)//')'
         ENDIF
      ENDIF

      WRITE(cval(11:80),sfmt) rval


C --- Verify that the length of the string written (after the removal of
C     multiple trailing zeros, if present, if 'cfmt' was blank on input)
C     is <= 20 bytes, and if so, right-justify it on byte 30 in header
C     record 'crec'...
      i = lenc(cval)
      j = clen(cval)

C     If this routine was called with a blank 'cfmt', ensure that it is
C     again blank upon exit.  In this case, also strip multiple trailing
C     zeros, leaving just one...
      IF (k.EQ.0) THEN
         CALL cinit(cfmt)

1        cc = cval(j:j)
         IF (cc.EQ.'0') THEN
            cval(j:j) = ' '
            j = j - 1
         ENDIF
         IF (cval(j:j).EQ.'0') GOTO 1
            
         j = j + 1
         cval(j:j) = cc
      ENDIF

C     Verify that 'cval' is <= 20 bytes...
      k = j - i + 1
      IF (k.GT.20) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +     'specified format exceeds 20-byte limit of FITS value field'
         ierr = 2396
         GOTO 9
      ENDIF

C     Right-justify the value in 'cval' on byte 30 in 'crec'...
      crec((30-k+1):30) = cval(i:j)
      crec(31:31) = ' '
      CALL cinit(cval)
      cval(11:80) = crec(11:80)


C --- Test that the string written has only blanks, digits, '+' or '-',
C     decimal point, or an "E" in bytes 11--30 ...
      DO i=11,30
         cc = cval(i:i)
         ic = ICHAR(cc)
         IF (     ic.NE.ib.AND.ic.NE.id
     +       .AND.ic.NE.ip.AND.ic.NE.im
     +       .AND.ic.NE.iEE.AND.(ic.LT.i0.OR.ic.GT.i9) ) THEN
C           Write appears to have failed (likely ***, or some such)..
            IF (idbg.GT.0) WRITE(*,*) dbid//
     +         'failed to write REAL keyword value to record.'
            ierr = 2397
            GOTO 9
         ENDIF
      ENDDO


C --- Try to read back the value written and compare with the original..
      rtmp = 0.
      READ(cval,*,IOSTAT=ierr) rtmp
      IF (ierr.NE.0) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//
     +      'failed to read back REAL keyword value from record.'
         ierr = 2398
         GOTO 9
      ENDIF
      rdev = ABS(rtmp-rval)/rval
      IF (idbg.GT.1) WRITE(*,*) dbid//
     +   ' in:',rval,', out:',rtmp,', error:',rdev
      IF (1.0d-7.LT.rdev) THEN
         IF (idbg.GT.0) WRITE(*,*) dbid//'keyword value '//
     +      'read back from record does not match input value.'
         ierr = 2399
         GOTO 9
      ENDIF


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


