C ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE rjfpkvals (crec, sval, lold, idbg, ierr)

C @(#)Subroutine: rjfpkvals          Author: R.A. Jansen -- Sep 13, 2009
C @(#)
C @(#)  Purpose: Write a character string-valued keyword value to a FITS
C @(#)    header record 'crec' for a keyword value 'sval' that is decla-
C @(#)    red as a variable  of the CHARACTER data type and of arbitrary
C @(#)    length ( CHARACTER*(*) internally).  The keyword value will be
C @(#)    written left-justified on byte 12 in the header record, prece-
C @(#)    ded and followed by a single quote. Embedded single quotes are
C @(#)    replaced with pairs of single quotes.  If the keyword value is
C @(#)    shorter than 8 characters and boolean switch  'lold'  is true,
C @(#)    then the keyword value  is padded with blanks  to 8 characters
C @(#)    before writing the closing quote. Padding is optional in v3.0,
C @(#)    but was mandatory  in previous versions  of the FITS standard.
C @(#)    If the keyword value is longer than the 68 characters that can
C @(#)    be accomodated  within a header record,  it will be truncated,
C @(#)    the closing quote is written in byte 80,  and an error will be
C @(#)    issued.   If verbosity level 'idbg'>0, additional messages may
C @(#)    be printed.  Error status 'ierr' will be non-zero on exit,  if
C @(#)    any error is detected.  Its value  (if >2000) uniquely identi-
C @(#)    fies a known error condition.
C @(#)
C @(#)  Calling sequence:
C @(#)    CALL rjfpkvals(crec, sval, lold, idbg, ierr),  where 'crec' is
C @(#)    a char string with a declared length of 80 chars, and 'sval' a
C @(#)    char string of arbitrary declared length.   Switch 'lold' is a
C @(#)    boolean, while 'idbg' and 'ierr' are both integers.
C @(#)
C @(#)  Note: this task is meant to write a char string keyword value to
C @(#)    a header record  that has been initialized already with a FITS
C @(#)    keyword;  use one of 'rjfmkval?' to modify an existing keyword
C @(#)    value.
C @(#)  Note2:  if 'sval' starts and/or ends with a single quote,  it is
C @(#)    assumed that it is significant  (a litteral single quote to be
C @(#)    preserved and thus represented by a pair of quotes). This will
C @(#)    result in 3 successive single quotes.
C @(#)
C @(#)  Dependencies:	clen, cinit
C @(#)
C @(#)  Revision history:
C @(#)    Aug 26, 2009 [RAJ] -- Initial version of this routine.
C @(#)    Sep  9, 2009 [RAJ] -- Made padding strings  to a minimum of  8
C @(#)                          characters as required by the  FITS v2.0
C @(#)                          and earlier standards optional by adding
C @(#)                          an 'lold' switch.
C @(#)    Sep 13, 2009 [RAJ] -- Changed truncation to be an error, while
C @(#)                          allowing the string value  to be written
C @(#)                          anyway.  Leave it to the calling routine
C @(#)                          to decide whether to continue on error.
C @(#)------------------------------------------------------------------

      IMPLICIT NONE

C --- Declaration of input/return variables...
      CHARACTER	crec*80, sval*(*)
      INTEGER	idbg, ierr
      LOGICAL	lold

C --- Declaration of local variables...
      CHARACTER dbid*19, cval*80, cc
      INTEGER	clen, ic, i, j, k, m

C     'ib', 'iq' and 'it' are shorthand integer equivalents in the ASCII
C     collating sequence of the space, quote and tilde, respectively...
      INTEGER	ib, iq, it
      CHARACTER	cq

C     ------------------------------------------------------------------

C --- Return for non-zero input error status...
      IF (ierr.NE.0) RETURN

C --- Initialize the debugging ID string...
      dbid  = 'DEBUG(rjfpkvals): '

C --- Initialize local variables...
      CALL cinit(cval)
      ib = 32
      iq = 39
      it = 126
      cq = CHAR(iq)


C --- Determine the minimum of the actual length of the input character
C     string and the 68 character maximum length that would fit (in the
C     absence of any embedded single quotes, that is)...
      i = clen(sval)
      j = MIN(i,68)
      IF (j.LT.i) THEN
         IF (idbg.GT.1) WRITE(*,*)
     +      dbid//' input character string will be truncated'
         ierr = 2341
      ENDIF
      IF (j.EQ.0.AND.idbg.GT.1) WRITE(*,*)
     +    dbid//' Warning: input character string is the null string'


C --- Try and write the char string-valued keyword value to character
C     string 'cval'; test on-the-fly for any illegal characters outside
C     the restricted ASCII set and replace any embedded single quotes
C     by pairs of single quotes...
      k = 0

C     Write the opening quote in byte 11...
      cval(11:11) = cq

C     The following loop must not be executed unless j>0 ...
      DO i=1,j
         cc = sval(i:i)
         ic = ICHAR(cc)
         IF (ic.LT.ib.OR.ic.GT.it) THEN
            IF (idbg.GT.0) WRITE(*,*)
     +         dbid//'illegal character within the keyword value.'
            IF (ierr.NE.0) THEN
               ierr = 2342
            ELSE
               ierr = 2343
            ENDIF
            GOTO 9
         ELSE IF (ic.EQ.iq) THEN
            IF (idbg.GT.2) WRITE(*,*)
     +         dbid//'  replaced embedded quote with pair of quotes.'
            m = 11 + k + i
            IF (m.LT.79) cval(m:m) = cq
            k = k + 1
         ENDIF
         m = 11 + k + i
         IF (m.LT.80) cval(m:m) = cc
      ENDDO

C     Compute the byte to which to write the closing quote...
      i = k + j
      IF (i.GT.68) THEN
         IF (j.LE.68) THEN
            IF (idbg.GT.1) WRITE(*,*) dbid//
     +      ' input string had to be truncated due to embedded quote'
            ierr = 2344
C           Although an error, continue anyway and let the calling
C           program decide what to do...
         ENDIF
         m = 80
      ELSE IF (i.LT.8.AND.lold) THEN
         IF (idbg.GT.2) WRITE(*,*) dbid//
     +      '  padded input string to minimum length of 8 characters'
         m = 20
      ELSE
         m = 12 + i
      ENDIF

C     Append the closing quote...
      cval(m:m) = cq


C --- Copy characters 11 through 80 to the output header record...
      crec(11:80) = cval(11:80)


9     IF (idbg.GT.2) WRITE(*,*) dbid//crec(1:clen(crec))
      IF (idbg.GT.1) WRITE(*,*) dbid//'on exit, ierr = ',ierr

      RETURN
      END


